/****************************************************************************/
/**    GNOM - Gnom Is NOt MCuT                                             **/
/**  Copyright (C) 1997 Jochen Sarrazin, DG6VJ. All righs reserved.        **/
/**                                                                        **/
/**  Routinen frs erweiterte AutoBIN-Protokoll                            **/
/****************************************************************************/


#include "global.h"

#include "utils.h"
#include "toolbox.h"
#include "sndqueue.h"
#include "display.h"

#include <string.h>


/* Welches Zeichen wird gestufft? Hier isses ascii 130. */
#define STUFF_CHAR '\x82'
/* Blckgre  -  x * 256 Bytes- 8 = 2k */
#define BLOCKSIZE 8

/* Blocktypen */
#define BLOCKTYPE_DATA 0x01
#define BLOCKTYPE_LAST 0x02
#define BLOCKTYPE_FIN 0x03
#define BLOCKTYPE_REQ 0xEF

#define EBIN_EOB "\xE6\x82\xE6"
#define EBIN_EOB_LEN 3

static int stuff_it(char *src, int slen, char *dest)
{
  int i=0,k=0;
  int found130=FALSE;
  

  while (i < slen)
  {
    if (src[i] == STUFF_CHAR)
    {
      if (!found130)
      {
        dest[k] = STUFF_CHAR;
        k++;
        found130 = TRUE;
      }
      dest[k] = STUFF_CHAR;
      k++;
    }
    else
    {
      found130 = FALSE;
      dest[k] = src[i];
      k++;
    }
    
    i++;
  }
  
  dest[k] = '\0';
  return(k);
}



static int unstuff_it(char *src, int slen, char *dest)
{
  int i=0,k=0;
  int found130=FALSE;
  
  
  while (i < slen)
  {
    if (src[i] == STUFF_CHAR)
    {
      if (found130)
      {
        dest[k] = STUFF_CHAR;
        k++;
      }
      found130 = TRUE;
    }
    else
    {
      found130 = FALSE;
      dest[k] = src[i];
      k++;
    }
    
    i++;
  }
  
  dest[k] = '\0';
  return(k);
}



/* Holt den nchsten Block frs extendet AutoBIN-Protokoll 'raus.
 * Das File mu geffnet sein, und der fd richtig bergeben.
 */
void ebin_get_block(int chan, int *fd)
{
  char *tmp,*str;
  int count,len,start_eob = FALSE;
  unsigned short crc,blocknr,blocklen;
  unsigned char blocktype;
  off_t offset;
  
  
  tmp = (char *) malloc(chaninfo[chan]->AutoEBINblocksize*256*2);
  str = (char *) malloc((chaninfo[chan]->AutoEBINblocksize*256)+50);
  
  /* Zum Errechnen der aktuellen Block-Nummer den Offset dieses Blocks im File holen */
  offset = lseek(*fd,0,SEEK_CUR);
  lseek(*fd,chaninfo[chan]->AutoEBINoffset,SEEK_SET);
  
  count = read(*fd,tmp,chaninfo[chan]->AutoEBINblocksize*256);
  chaninfo[chan]->AutoEBINoffset += count;
  
  chaninfo[chan]->AutoEBIN = (char *)malloc(count+50);
  
  /* Block-Type */
  if (count == 0)
     blocktype = BLOCKTYPE_FIN;
  else if (count != BLOCKSIZE*256)
          blocktype = BLOCKTYPE_LAST;
       else
          blocktype = BLOCKTYPE_DATA;
  
  /* Blocktypen nach str schreiben */
  memcpy(str,&blocktype,1);
  
  switch (blocktype)
  {
    case BLOCKTYPE_FIN:
         /* FIN-Block */
         memcpy(str+1,&chaninfo[chan]->AutoEBINcrc,4);
    
         len = 1;
         chaninfo[chan]->AutoEBINready = TRUE;
         break;
         
    case BLOCKTYPE_DATA:
         memcpy(str+1,&chaninfo[chan]->AutoEBINblocksize,1);
    
         /* Blocknummer errechnen */
         blocknr = offset / (chaninfo[chan]->AutoEBINblocksize * 256);
    
         memcpy(str+2,&blocknr,2);
    
         /* Jetzt kommen die Daten */
         memcpy(str+4,tmp,count);
    
         len = 4+count;
    
         /* Wenn das der erste Block ist, wird ein EOB-Paket an den Anfang gestellt */
         if (offset == 0) start_eob = TRUE;
         break;
         
    case BLOCKTYPE_LAST:
         /* Genauer block-offset */
         memcpy(str+1,&offset,4);
    
         /* genaue Block-Lnge */
         blocklen = count;
         memcpy(str+5,&blocklen,2);
    
         /* Jetzt kommen die Daten */
         memcpy(str+7,tmp,count);
    
         len = 7+count;
    
         if (offset == 0) start_eob = TRUE;
         break;
  }
  
  
  /* CRC fr diesen Block errechnen und anhngen */
  crc = calc_crcthp_str(str,len,0);
  memcpy(str+len,&crc,2);
  len += 2;
  
  /* Stuffen und nach AutoEBIN schreiben */
  chaninfo[chan]->AutoEBINlen = stuff_it(str,len,chaninfo[chan]->AutoEBIN);
  
  /* EOB-Markierung */
  memcpy(chaninfo[chan]->AutoEBIN+chaninfo[chan]->AutoEBINlen,EBIN_EOB,EBIN_EOB_LEN);
  chaninfo[chan]->AutoEBINlen += 3;
  
  /* Wenn dieser Block der erste ist, wird davor auch noch eine EOB-Markierung gesetzt. */
  if (start_eob)
  {
    memmove(chaninfo[chan]->AutoEBIN+EBIN_EOB_LEN,chaninfo[chan]->AutoEBIN,chaninfo[chan]->AutoEBINlen);
    chaninfo[chan]->AutoEBINlen += EBIN_EOB_LEN;
    memcpy(chaninfo[chan]->AutoEBIN,EBIN_EOB,EBIN_EOB_LEN);
  }
  
  /****************************** DEBUGGING ****************************************************/
  if ((chaninfo[chan]->AutoEBIN[0]==1) && (chaninfo[chan]->AutoEBIN[2]==2))
  {
    memcpy(chaninfo[chan]->AutoEBIN+20,"XX",2);
  }
  /****************************** DEBUGGING ****************************************************/
  
  free(tmp);
  free(str);
}



void ebin_send_request_answer(int chan, unsigned long offset, unsigned long blocklen)
{
  int len,slen;
  char *tmp,*str;
  unsigned short crc;
  unsigned long of;
  
  
  of = lseek(chaninfo[chan]->autobin->fd,offset,SEEK_SET);
  
  tmp = (char *)malloc(blocklen+50);
  read(chaninfo[chan]->autobin->fd,tmp+7,blocklen);
  
  
  /* Blocktypen nach str schreiben */
  tmp[0] = BLOCKTYPE_LAST;
  
  /* Genauer block-offset */
  memcpy(tmp+1,&offset,4);
  
  /* genaue Block-Lnge */
  memcpy(tmp+5,&blocklen,2);
  
  len = blocklen+7;
  
  
  /* CRC fr diesen Block errechnen und anhngen */
  crc = calc_crcthp_str(tmp,len,0);
  memcpy(tmp+len,&crc,2);
  len += 2;
  
  /* Stuffen und nach AutoEBIN schreiben */
  str = (char *)malloc(len*2);
  slen = stuff_it(tmp,len,str);
  free(tmp);
  
  /* EOB-Markierung */
  memcpy(str+slen,EBIN_EOB,EBIN_EOB_LEN);
  slen += 3;
  
  send_Bin(chan,str,slen);
}



/* Irgendwelche EBIN-Befehle? (#REQ# oder sonstiges) */
void ebin_commands(int chan, char *data, int len)
{
  char *tmp,*str;
  int tmplen,slen,pos;
  unsigned long offset=0, blocklen=0;
  unsigned short crc,ourcrc;
  
  
  if (chaninfo[chan]->ebin == NULL)
  {
    chaninfo[chan]->ebin = (s_ebin *)malloc(sizeof(s_ebin));
    chaninfo[chan]->ebin->len = 0;
    chaninfo[chan]->ebin->str = NULL;
  }
  
  /* Den empfangenen String an den schon frher empfangenen anhngen */
  if (chaninfo[chan]->ebin->len == 0)
  {
    chaninfo[chan]->ebin->str = (char *)malloc(len);
    memcpy(chaninfo[chan]->ebin->str,data,len);
    chaninfo[chan]->ebin->len = len;
  }
  else
  {
    tmp = chaninfo[chan]->ebin->str;
    tmplen = chaninfo[chan]->ebin->len;
    chaninfo[chan]->ebin->str = (char *)malloc(tmplen+len);
    memcpy(chaninfo[chan]->ebin->str,tmp,tmplen);
    memcpy(chaninfo[chan]->ebin->str+tmplen,data,len);
    chaninfo[chan]->ebin->len = tmplen + len;
    free(tmp);
  }
  
  if ((pos = memstr(chaninfo[chan]->ebin->str,EBIN_EOB,chaninfo[chan]->ebin->len)) == -1) return;
  
  tmp = chaninfo[chan]->ebin->str;
  tmplen = chaninfo[chan]->ebin->len;
  
  chaninfo[chan]->ebin->str = (char *)malloc(chaninfo[chan]->ebin->len-pos-EBIN_EOB_LEN);
  chaninfo[chan]->ebin->len = chaninfo[chan]->ebin->len - pos - EBIN_EOB_LEN;
  memcpy(chaninfo[chan]->ebin->str,chaninfo[chan]->ebin->str+pos+EBIN_EOB_LEN,chaninfo[chan]->ebin->len);
  
  tmplen = pos;
  
  str = (char *)malloc(tmplen*2);
  slen = unstuff_it(tmp,tmplen,str);
  free(tmp);
  
  switch ((uchar)str[0])
  {
    case BLOCKTYPE_REQ:
         memcpy(&offset,str+1,4);
         memcpy(&blocklen,str+5,4);
         memcpy(&crc,str+9,2);
         if (slen == 11)
         {
           /* Kein Byte ist verloren gegangen - CRC prfen */
           ourcrc = calc_crcthp_str(str,9,0);
         }
         else ourcrc = crc+1;

         if (crc == ourcrc)
            ebin_send_request_answer(chan, offset, blocklen);
         
         break;
  }
}



void ebin_check_ebinline(char *data, int chan, int len)
{
  char *str,tmp[500],*filename,namelen,*fullname;
  unsigned long filelen=-1, crc32=-1;
  off_t offset=-1;
  unsigned short crc16=-1;
  time_t epochtime=-1;
  unsigned char filetype;
  int i,fd,fd2;
  struct utimbuf utimetmp;
  
  
  
  len -= 5;
  
  str = (char *)malloc(len+1);
  memcpy(str,data+5,len);
  
  while (len > 0)
  {
    switch (str[1])
    {
      case '0': /* File-Lnge (in hex) */
                memcpy(tmp,str+2,4);
                tmp[4] = '\0';
                len -= 6;
                memcpy(str,str+6,len);
                filelen = get_hex(tmp);
                break;
      case '1': /* Filenamen */
                memcpy(tmp,str+2,2);
                tmp[2] = '\0';
                namelen = get_hex(tmp);
                filename = (char *)malloc(namelen+1);
                memcpy(filename,str+4,namelen);
                filename[namelen] = '\0';
                len -= (4+namelen);
                memcpy(str,str+4+namelen,len);
                break;
      case '2': /* 16-Bit-CRC */
                memcpy(tmp,str+2,4);
                tmp[4] = '\0';
                len -= 6;
                memcpy(str,str+6,len);
                crc16 = get_hex(tmp);
                break;
      case '3': /* 32-Bit-CRC */
                memcpy(tmp,str+2,8);
                tmp[8] = '\0';
                len -= 10;
                memcpy(str,str+10,len);
                crc32 = get_hex(tmp);
                break;
      case '4': /* epoch-time */
                memcpy(tmp,str+2,8);
                tmp[8] = '\0';
                len -= 10;
                memcpy(str,str+10,len);
                epochtime = get_hex(tmp);
                break;
      case '5': /* Filetype */
                filetype = str[2];
                break;
    }
  }
  
  free(str);
  
  if (filelen == 0)
  {
    send_Textline(chan,"#NO# - invalid filelength\xD");
    free(filename);
    return;
  }
  
  
  /* Ein Dateiname ist vorhanden - einen evtl. vorhandenen Pfad lschen */
  fullname = (char *)strdup(filename);
  while ((i = POS('/',filename)) != -1) COPY(filename,filename,i+1,strlen(filename)-i-1); 
  while ((i = POS('\\',filename)) != -1) COPY(filename,filename,i+1,strlen(filename)-i-1);
  
  sprintf(tmp,"%s/abin/%s",config->maindir,filename);
  
  if ((fd = open(tmp,O_RDWR)) == -1)
     if ((fd = open(tmp,O_RDWR|O_CREAT|O_TRUNC)) == -1)
     {
       send_Textline(chan,"#NO# - cannot open destination-file\xD");
       free(filename);
       free(fullname);
       return;
     }
  
  /* Zugriffsrechte einstellen */
  fchmod(fd,S_IRUSR|S_IWUSR);
  
  sprintf(tmp,"%s/abin/%s.resume",config->maindir,filename);
  
  /* Gucken, ob so ein Resume-File vorhanden ist. */
  if ((fd2 = open(tmp,O_RDWR)) != -1)
     offset = lseek(fd2,0,SEEK_END);
  else
     if ((fd2 = open(tmp,O_RDWR|O_CREAT|O_TRUNC)) == -1)
     {
       send_Textline(chan,"#NO# - cannot open resume-file\xD");
       free(filename);
       free(fullname);
       return;
     }
  
  /* Zugriffsrechte einstellen */
  fchmod(fd2,S_IRUSR|S_IWUSR);
  
  
  /* Datum + Uhrzeit einstellen */
  if (epochtime != 0)
  {
    utimetmp.actime = epochtime;
    utimetmp.modtime = epochtime;
    utime(tmp,&utimetmp);
  }
  
  /* Jetzt chaninfo[]->autobin allokieren und fllen */
  chaninfo[chan]->autobin = (s_autobin *)malloc(sizeof(s_autobin));
  chaninfo[chan]->autobin->len = filelen;
  chaninfo[chan]->autobin->bytes_got = 0;
  chaninfo[chan]->autobin->crc_rx = crc16;
  chaninfo[chan]->autobin->crc_tx = 0;
  chaninfo[chan]->autobin->crc32 = crc32;
  chaninfo[chan]->autobin->calc_crc = TRUE;
  chaninfo[chan]->autobin->type = 2;
  
  chaninfo[chan]->autobin->fd = fd;
  chaninfo[chan]->autobin->markfd = fd2;
  chaninfo[chan]->autobin->name = (char *)strdup(filename);
  chaninfo[chan]->autobin->fullname = fullname;
  chaninfo[chan]->flags &= ~CH_LINEMODE;
  
  chaninfo[chan]->flags |= CH_AUTOBINRX;
  if ((chaninfo[chan]->flags & CH_SAVE) == CH_SAVE)
  {
    chaninfo[chan]->oldflags |= CH_SAVE;
    chaninfo[chan]->flags &= ~CH_SAVE;
  }
  if (chan == AktChan) update_upper_statusline(TRUE);
  
  if (offset == -1)
     send_Textline(chan,"#EOK#\xD");
  else
  {
    sprintf(tmp,"#EOK#%i\xD",(int)offset);
    send_Textline(chan,tmp);
  }
  
  chaninfo[chan]->autobin->starttime = time(NULL);
  
  
  chaninfo[chan]->ebin = (s_ebin *)malloc(sizeof(s_ebin));
  chaninfo[chan]->ebin->save_it = FALSE;
  chaninfo[chan]->ebin->str = NULL;
  chaninfo[chan]->AutoEBINblocksize = BLOCKSIZE;

  
  free(filename);
}



void ebin_add_data(int fd, char *str, unsigned long len, unsigned long offs)
{
  lseek(fd,offs,SEEK_SET);
  write(fd,str,len);
}



void ebin_add_dataflag(int fd, unsigned long len, unsigned long offs)
{
  char *tmp;
  
  tmp = (char *)malloc(len+1);
  StringD(tmp,len,'X');
  
  lseek(fd,offs+20,SEEK_SET);
  write(fd,tmp,len);
}



void ebin_send_req(int chan, unsigned long offset, unsigned long bytecount)
{
  char tmp[50],str[100];
  int len,slen;
  unsigned short crc;
  
  tmp[0] = (char)0xEF;
  memcpy(tmp+1,&offset,4);
  memcpy(tmp+5,&bytecount,4);
  len = 9;
  
  /* CRC fr diesen Block errechnen und anhngen */
  crc = calc_crcthp_str(tmp,len,0);
  memcpy(tmp+len,&crc,2);
  len += 2;
  
  /* Stuffen und nach AutoEBIN schreiben */
  slen = stuff_it(tmp,len,str);
  
  /* EOB-Markierung */
  memcpy(str+slen,EBIN_EOB,EBIN_EOB_LEN);
  slen += 3;
  
  send_Bin(chan, str, slen);
}



void ebin_saveframe(char *data, int len, int chan)
{
  char *tmp,*str;
  int tmplen,pos,slen;
  unsigned char blocksize;
  unsigned short blocknr;
  unsigned short ourcrc,crc,blocklen;
  off_t offset;

  /* Den empfangenen String an den schon frher empfangenen anhngen */
  if (chaninfo[chan]->ebin->len == 0)
  {
    chaninfo[chan]->ebin->str = (char *)malloc(len);
    memcpy(chaninfo[chan]->ebin->str,data,len);
    chaninfo[chan]->ebin->len = len;
  }
  else
  {
    tmp = chaninfo[chan]->ebin->str;
    tmplen = chaninfo[chan]->ebin->len;
    chaninfo[chan]->ebin->str = (char *)malloc(tmplen+len);
    memcpy(chaninfo[chan]->ebin->str,tmp,tmplen);
    memcpy(chaninfo[chan]->ebin->str+tmplen,data,len);
    chaninfo[chan]->ebin->len = tmplen + len;
    free(tmp);
  }


  if (!chaninfo[chan]->ebin->save_it)
  {
    /* Noch nicht abspeichern - erstmal auf einen EOB-Block warten */
    if ((pos = memstr(chaninfo[chan]->ebin->str,EBIN_EOB,chaninfo[chan]->ebin->len)) == -1) return;
    
    pos += EBIN_EOB_LEN;
    memmove(chaninfo[chan]->ebin->str,chaninfo[chan]->ebin->str+pos,chaninfo[chan]->ebin->len-pos);
    chaninfo[chan]->ebin->len -= pos;
    
    chaninfo[chan]->ebin->save_it = TRUE;
  }

  if ((pos = memstr(chaninfo[chan]->ebin->str,EBIN_EOB,chaninfo[chan]->ebin->len)) == -1) return;

  tmp = chaninfo[chan]->ebin->str;
  tmplen = chaninfo[chan]->ebin->len;

  chaninfo[chan]->ebin->str = (char *)malloc(chaninfo[chan]->ebin->len-pos-EBIN_EOB_LEN);
  chaninfo[chan]->ebin->len = chaninfo[chan]->ebin->len - pos - EBIN_EOB_LEN;
  memcpy(chaninfo[chan]->ebin->str,chaninfo[chan]->ebin->str+pos+EBIN_EOB_LEN,chaninfo[chan]->ebin->len);

  tmplen = pos;

  str = (char *)malloc(tmplen*2);
  slen = unstuff_it(tmp,tmplen,str);
  free(tmp);

  switch (str[0])
  {
    case BLOCKTYPE_DATA:
         blocksize = str[1];
         memcpy(&blocknr,str+2,2);
         tmp = (char *)malloc(blocksize*256*2);
         if (slen == (blocksize*256)+6)
         {
           /* Kein Byte ist verloren gegangen - CRC prfen */
           memcpy(tmp,str,4+blocksize*256);
           memcpy(&crc,str+4+(blocksize*256),2);
           ourcrc = calc_crcthp_str(tmp,4+blocksize*256,0);
         }
         else ourcrc = crc+1;
         
         if (ourcrc != crc)
         {
           /* Neu beantragen */
           ebin_send_req(chan,blocknr*blocksize*256,blocksize*256);
         }
         else
         {
           /* Abspeichern */
           ebin_add_data(chaninfo[chan]->autobin->fd,tmp+4,blocksize*256,blocknr*blocksize*256);
           ebin_add_dataflag(chaninfo[chan]->autobin->markfd,blocksize*256,blocknr*blocksize*256);
           chaninfo[chan]->autobin->bytes_got += blocksize*256;
           update_lower_statusline(TRUE);
         }
         free(tmp);
         break;
    case BLOCKTYPE_LAST:
         memcpy(&offset,tmp+1,4);
         memcpy(&blocklen,tmp+5,2);
         tmp = (char *)malloc(blocklen*2);
         if (slen == blocklen+9)
         {
           /* Kein Byte ist verloren gegangen - CRC prfen */
           memcpy(tmp,str,7+blocklen);
           memcpy(&crc,str+7+blocklen,2);
           ourcrc = calc_crcthp_str(tmp,7+blocklen,0);
         }
         else ourcrc = crc+1;
         
         if (ourcrc != crc)
         {
           /* Neu beantragen */
           ebin_send_req(chan,offset,blocklen);
         }
         else
         {
           /* Abspeichern */
           ebin_add_data(chaninfo[chan]->autobin->fd,tmp+7,blocklen,offset);
           ebin_add_dataflag(chaninfo[chan]->autobin->markfd,blocklen,offset);
           chaninfo[chan]->autobin->bytes_got += blocklen;
           update_lower_statusline(TRUE);
         }
         free(tmp);
         break;
    case BLOCKTYPE_FIN:
         break;
  }
}



